/*********************************************************************
* Copyright (c) 2022 Eclipse Foundation.
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* Author: Martin Lowe <martin.lowe@eclipse-foundation.org>
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
package org.eclipsefoundation.git.eca.service.impl;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.Map.Entry;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import javax.ws.rs.WebApplicationException;

import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.eclipse.microprofile.rest.client.inject.RestClient;
import org.eclipsefoundation.core.service.CachingService;
import org.eclipsefoundation.git.eca.api.AccountsAPI;
import org.eclipsefoundation.git.eca.api.BotsAPI;
import org.eclipsefoundation.git.eca.model.EclipseUser;
import org.eclipsefoundation.git.eca.model.Project;
import org.eclipsefoundation.git.eca.service.OAuthService;
import org.eclipsefoundation.git.eca.service.UserService;
import org.jboss.resteasy.specimpl.MultivaluedMapImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.JsonNode;

/**
 * Wrapped cached and authenticated access to user objects.
 * 
 * @author Martin Lowe
 *
 */
@ApplicationScoped
public class CachedUserService implements UserService {
    private static final Logger LOGGER = LoggerFactory.getLogger(CachedUserService.class);

    @Inject
    @ConfigProperty(name = "eclipse.noreply.email-patterns")
    List<String> emailPatterns;

    // eclipse API rest client interfaces
    @Inject
    @RestClient
    AccountsAPI accounts;
    @Inject
    @RestClient
    BotsAPI bots;

    @Inject
    OAuthService oauth;
    @Inject
    CachingService cache;

    // rendered list of regex values
    List<Pattern> patterns;

    @PostConstruct
    void init() {
        // compile the patterns once per object to save processing time
        this.patterns = emailPatterns.stream().map(Pattern::compile).collect(Collectors.toList());
    }

    @Override
    public EclipseUser getUser(String mail) {
        Optional<EclipseUser> u = cache.get(mail, new MultivaluedMapImpl<>(), EclipseUser.class,
                () -> retrieveUser(mail));
        if (u.isPresent()) {
            LOGGER.debug("Found user with email {}", mail);
            return u.get();
        }
        LOGGER.debug("Could not find user with email {}", mail);
        return null;
    }

    @Override
    public EclipseUser getUserByGithubUsername(String username) {
        Optional<EclipseUser> u = cache.get("gh:" + username, new MultivaluedMapImpl<>(), EclipseUser.class,
                () -> accounts.getUserByGithubUname(getBearerToken(), username));
        if (u.isPresent()) {
            LOGGER.debug("Found user with name {}", username);
            return u.get();
        }
        LOGGER.debug("Could not find user with name {}", username);
        return null;
    }

    @Override
    public boolean userIsABot(String mail, List<Project> filteredProjects) {
        if (mail == null || "".equals(mail.trim())) {
            return false;
        }
        List<JsonNode> botObjs = getBots();
        // if there are no matching projects, then check against all bots, not just
        // project bots
        if (filteredProjects == null || filteredProjects.isEmpty()) {
            return botObjs.stream().anyMatch(bot -> checkFieldsForMatchingMail(bot, mail));
        }
        // for each of the matched projects, check the bot for the matching project ID
        for (Project p : filteredProjects) {
            LOGGER.debug("Checking project {} for matching bots", p.getProjectId());
            for (JsonNode bot : botObjs) {
                // if the project ID match, and one of the email fields matches, then user is
                // bot
                if (p.getProjectId().equalsIgnoreCase(bot.get("projectId").asText())
                        && checkFieldsForMatchingMail(bot, mail)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Checks for standard and noreply email address matches for a Git user and
     * converts to a Eclipse Foundation account object.
     * 
     * @param user the user to attempt account retrieval for.
     * @return the user account if found by mail, or null if none found.
     */
    private EclipseUser retrieveUser(String mail) {
        LOGGER.debug("Getting fresh user for {}", mail);
        // check for noreply (no reply will never have user account, and fails fast)
        EclipseUser eclipseUser = checkForNoReplyUser(mail);
        if (eclipseUser != null) {
            return eclipseUser;
        }
        // standard user check (returns best match)
        LOGGER.debug("Checking user with mail {}", mail);
        try {
            List<EclipseUser> matches = accounts.getUsers(getBearerToken(), mail);
            if (matches != null && !matches.isEmpty()) {
                return matches.get(0);
            }
        } catch (WebApplicationException e) {
            LOGGER.warn("Could not find user account with mail '{}'", mail);
        }
        return null;
    }

    /**
     * Checks git user for no-reply address, and attempts to ratify user through
     * reverse lookup in API service. Currently, this service only recognizes Github
     * no-reply addresses as they have a route to be mapped.
     * 
     * @param user the Git user account to check for no-reply mail address
     * @return the Eclipse user if email address is detected no reply and one can be
     *         mapped, otherwise null
     */
    private EclipseUser checkForNoReplyUser(String mail) {
        LOGGER.debug("Checking user with mail {} for no-reply", mail);
        boolean isNoReply = patterns.stream().anyMatch(pattern -> pattern.matcher(mail.trim()).find());
        if (isNoReply) {
            // get the username/ID string before the first @ symbol.
            String noReplyUser = mail.substring(0, mail.indexOf("@", 0));
            // split based on +, if more than one part, use second (contains user),
            // otherwise, use whole string
            String[] nameParts = noReplyUser.split("[\\+]");
            String namePart;
            if (nameParts.length > 1 && nameParts[1] != null) {
                namePart = nameParts[1];
            } else {
                namePart = nameParts[0];
            }
            String uname = namePart.trim();
            LOGGER.debug("User with mail {} detected as noreply account, checking services for username match on '{}'",
                    mail, uname);

            // check github for no-reply (only allowed noreply currently)
            if (mail.endsWith("noreply.github.com")) {
                try {
                    // check for Github no reply, return if set
                    return getUserByGithubUsername(uname);
                } catch (WebApplicationException e) {
                    LOGGER.warn("No match for '{}' in Github", uname);
                }
            }
        }
        return null;
    }

    /**
     * Checks JSON node to look for email fields, both at the root, and nested email
     * fields.
     * 
     * @param bot  the bots JSON object representation
     * @param mail the email to match against
     * @return true if the bot has a matching email value, otherwise false
     */
    private boolean checkFieldsForMatchingMail(JsonNode bot, String mail) {
        // check the root email for the bot for match
        JsonNode botmail = bot.get("email");
        if (mail != null && botmail != null && mail.equalsIgnoreCase(botmail.asText(""))) {
            LOGGER.debug("Found matching bot at root level for '{}'", mail);
            return true;
        }
        Iterator<Entry<String, JsonNode>> i = bot.fields();
        while (i.hasNext()) {
            Entry<String, JsonNode> e = i.next();
            // check that our field is an object with fields
            JsonNode node = e.getValue();
            if (node.isObject()) {
                LOGGER.debug("Checking {} for bot email", e.getKey());
                // if the mail matches (ignoring case) user is bot
                JsonNode botAliasMail = node.get("email");
                if (mail != null && botAliasMail != null && mail.equalsIgnoreCase(botAliasMail.asText(""))) {
                    LOGGER.debug("Found match for bot email {}", mail);
                    return true;
                }
            }
        }
        return false;
    }

    private List<JsonNode> getBots() {
        Optional<List<JsonNode>> allBots = cache.get("allBots", new MultivaluedMapImpl<>(), JsonNode.class,
                () -> bots.getBots());
        if (!allBots.isPresent()) {
            return Collections.emptyList();
        }
        return allBots.get();
    }

    private String getBearerToken() {
        return "Bearer " + oauth.getToken();
    }

}
