/*********************************************************************
* Copyright (c) 2020 Eclipse Foundation.
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* Author: Martin Lowe <martin.lowe@eclipse-foundation.org>
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
package org.eclipsefoundation.git.eca.helper;

import java.util.ArrayList;

import org.eclipsefoundation.git.eca.model.Commit;
import org.eclipsefoundation.git.eca.model.GitUser;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import io.quarkus.test.junit.QuarkusTest;

/**
 * Tests related to the {@linkplain CommitHelper} class.
 * 
 * @author Martin Lowe
 *
 */
@QuarkusTest
class CommitHelperTest {

    // represents a known good commit before the start of each test
    GitUser testUser;
    Commit.Builder baseCommit;

    @BeforeEach
    void setup() {
        // basic good user
        testUser = GitUser.builder().setMail("test.user@eclipse-foundation.org").setName("Tester McTesterson").build();

        // basic known good commit
        baseCommit = Commit.builder()
                .setBody(String.format("Sample body content\n\nSigned-off-by: %s <%s>", testUser.getName(),
                        testUser.getMail()))
                .setHash("abc123f").setHead(false).setParents(new ArrayList<>())
                .setSubject("Testing CommitHelper class #1337").setAuthor(testUser).setCommitter(testUser);
    }

    @Test
    void validateCommitKnownGood() {
        Assertions.assertTrue(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to pass validation");
    }

    @Test
    void validateCommitNullCommit() {
        Assertions.assertFalse(CommitHelper.validateCommit(null), "Expected null commit to fail validation");
    }

    @Test
    void validateCommitNoAuthor() {
        baseCommit.setAuthor(null);
        Assertions.assertFalse(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to fail validation w/ no author");
    }

    @Test
    void validateCommitNoAuthorMail() {
        baseCommit.setAuthor(GitUser.builder().setName("Some Name").build());
        Assertions.assertFalse(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to fail validation w/ no author mail address");
    }

    @Test
    void validateCommitNoCommitter() {
        baseCommit.setCommitter(null);
        Assertions.assertFalse(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to fail validation w/ no committer");
    }

    @Test
    void validateCommitNoCommitterMail() {
        baseCommit.setCommitter(GitUser.builder().setName("Some Name").build());
        Assertions.assertFalse(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to fail validation w/ no committer mail address");
    }

    @Test
    void validateCommitNoHash() {
        baseCommit.setHash(null);
        Assertions.assertFalse(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to fail validation w/ no commit hash");
    }

    @Test
    void validateCommitNoBody() {
        baseCommit.setBody(null);
        Assertions.assertTrue(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to pass validation w/ no body");
    }

    @Test
    void validateCommitNoParents() {
        baseCommit.setParents(new ArrayList<>());
        Assertions.assertTrue(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to pass validation w/ no parents");
    }

    @Test
    void validateCommitNoSubject() {
        baseCommit.setSubject(null);
        Assertions.assertTrue(CommitHelper.validateCommit(baseCommit.build()),
                "Expected basic commit to pass validation w/ no subject");
    }
}
